from maverick import Maverick

a = Maverick()
# r = a.predict("Hi i am Giuliano, i like pasta. Giuliano, my cool name")
print(
    a.predict(
        [
            [
                "--",
                "basically",
                ",",
                "it",
                "was",
                "unanimously",
                "agreed",
                "upon",
                "by",
                "the",
                "various",
                "relevant",
                "parties",
                ".",
            ],
            [
                "To",
                "express",
                "its",
                "determination",
                ",",
                "the",
                "Chinese",
                "securities",
                "regulatory",
                "department",
                "compares",
                "this",
                "stock",
                "reform",
                "to",
                "a",
                "die",
                "that",
                "has",
                "been",
                "cast",
                ".",
            ],
            [
                "It",
                "takes",
                "time",
                "to",
                "prove",
                "whether",
                "the",
                "stock",
                "reform",
                "can",
                "really",
                "meet",
                "expectations",
                ",",
                "and",
                "whether",
                "any",
                "deviations",
                "that",
                "arise",
                "during",
                "the",
                "stock",
                "reform",
                "can",
                "be",
                "promptly",
                "corrected",
                ".",
            ],
            ["Dear", "viewers", ",", "the", "China", "News", "program", "will", "end", "here", "."],
            ["This", "is", "Xu", "Li", "."],
            ["Thank", "you", "everyone", "for", "watching", "."],
            ["Coming", "up", "is", "the", "Focus", "Today", "program", "hosted", "by", "Wang", "Shilin", "."],
            ["Good-bye", ",", "dear", "viewers", "."],
        ],
        # predefined_mentions=[[16, 16], [19, 23], [42, 44], [57, 59], [25, 27], [83, 83], [82, 82]],
        add_gold_clusters=[[[1, 1], [2, 2]]],
        singletons=False,
    )
)
